/*
 * Copyright (c) 2022-2024 Philippe Proulx <pproulx@efficios.com>
 *
 * SPDX-License-Identifier: MIT
 */

#include "json-strings.hpp"

namespace ctf {
namespace jsonstr {

const char * const accuracy = "accuracy";
const char * const align = "alignment";
const char * const attrs = "attributes";
const char * const bigEndian = "big-endian";
const char * const bitOrder = "bit-order";
const char * const btNs = "babeltrace.org,2020";
const char * const byteOrder = "byte-order";
const char * const clkCls = "clock-class";
const char * const cycles = "cycles";
const char * const dataStreamCls = "data-stream-class";
const char * const dataStreamClsId = "data-stream-class-id";
const char * const dataStreamId = "data-stream-id";
const char * const defClkClsId = "default-clock-class-id";
const char * const defClkTs = "default-clock-timestamp";
const char * const descr = "description";
const char * const discEventRecordCounterSnap = "discarded-event-record-counter-snapshot";
const char * const dynLenArray = "dynamic-length-array";
const char * const dynLenBlob = "dynamic-length-blob";
const char * const dynLenStr = "dynamic-length-string";
const char * const elemFc = "element-field-class";
const char * const emfUri = "emf-uri";
const char * const encoding = "encoding";
const char * const env = "environment";
const char * const eventRecordCls = "event-record-class";
const char * const eventRecordClsId = "event-record-class-id";
const char * const eventRecordCommonCtx = "event-record-common-context";
const char * const eventRecordCommonCtxFc = "event-record-common-context-field-class";
const char * const eventRecordHeader = "event-record-header";
const char * const eventRecordHeaderFc = "event-record-header-field-class";
const char * const eventRecordPayload = "event-record-payload";
const char * const eventRecordSpecCtx = "event-record-specific-context";
const char * const extensions = "extensions";
const char * const fc = "field-class";
const char * const fcAlias = "field-class-alias";
const char * const fixedLenBitArray = "fixed-length-bit-array";
const char * const fixedLenBitMap = "fixed-length-bit-map";
const char * const fixedLenBool = "fixed-length-boolean";
const char * const fixedLenFloat = "fixed-length-floating-point-number";
const char * const fixedLenSInt = "fixed-length-signed-integer";
const char * const fixedLenUInt = "fixed-length-unsigned-integer";
const char * const flags = "flags";
const char * const freq = "frequency";
const char * const ftl = "first-to-last";
const char * const id = "id";
const char * const len = "length";
const char * const lenFieldLoc = "length-field-location";
const char * const littleEndian = "little-endian";
const char * const logLevel = "log-level";
const char * const logLevelAlert = "alert";
const char * const logLevelCritical = "critical";
const char * const logLevelDebugFunction = "debug:function";
const char * const logLevelDebugLine = "debug:line";
const char * const logLevelDebugModule = "debug:module";
const char * const logLevelDebug = "debug";
const char * const logLevelDebugProcess = "debug:process";
const char * const logLevelDebugProgram = "debug:program";
const char * const logLevelDebugSystem = "debug:system";
const char * const logLevelDebugUnit = "debug:unit";
const char * const logLevelEmergency = "emergency";
const char * const logLevelError = "error";
const char * const logLevelInfo = "info";
const char * const logLevelNotice = "notice";
const char * const logLevelWarning = "warning";
const char * const ltf = "last-to-first";
const char * const mappings = "mappings";
const char * const mediaType = "media-type";
const char * const memberClasses = "member-classes";
const char * const metadataStreamUuid = "metadata-stream-uuid";
const char * const minAlign = "minimum-alignment";
const char * const name = "name";
const char * const ns = "namespace";
const char * const nullTerminatedStr = "null-terminated-string";
const char * const offsetFromOrigin = "offset-from-origin";
const char * const optional = "optional";
const char * const opts = "options";
const char * const origin = "origin";
const char * const path = "path";
const char * const payloadFc = "payload-field-class";
const char * const pktContentLen = "packet-content-length";
const char * const pktCtx = "packet-context";
const char * const pktCtxFc = "packet-context-field-class";
const char * const pktEndDefClkTs = "packet-end-default-clock-timestamp";
const char * const pktHeader = "packet-header";
const char * const pktHeaderFc = "packet-header-field-class";
const char * const pktMagicNumber = "packet-magic-number";
const char * const pktSeqNum = "packet-sequence-number";
const char * const pktTotalLen = "packet-total-length";
const char * const preamble = "preamble";
const char * const precision = "precision";
const char * const prefDispBase = "preferred-display-base";
const char * const roles = "roles";
const char * const seconds = "seconds";
const char * const selFieldLoc = "selector-field-location";
const char * const selFieldRanges = "selector-field-ranges";
const char * const specCtxFc = "specific-context-field-class";
const char * const staticLenArray = "static-length-array";
const char * const staticLenBlob = "static-length-blob";
const char * const staticLenStr = "static-length-string";
const char * const structure = "structure";
const char * const traceCls = "trace-class";
const char * const type = "type";
const char * const uid = "uid";
const char * const unixEpoch = "unix-epoch";
const char * const utf16Be = "utf-16be";
const char * const utf16Le = "utf-16le";
const char * const utf32Be = "utf-32be";
const char * const utf32Le = "utf-32le";
const char * const utf8 = "utf-8";
const char * const uuid = "uuid";
const char * const variant = "variant";
const char * const varLenSInt = "variable-length-signed-integer";
const char * const varLenUInt = "variable-length-unsigned-integer";
const char * const version = "version";

} /* namespace jsonstr */
} /* namespace ctf */
