/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright (c) 2017 Philippe Proulx <pproulx@efficios.com>
 */

#define BT_LOG_OUTPUT_LEVEL bt_lib_ctf_writer_log_level
#include "logging/log.h"

BT_LOG_INIT_LOG_LEVEL(bt_lib_ctf_writer_log_level,
	"BABELTRACE_CTF_WRITER_LOG_LEVEL");
