/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright 2013, 2014 Jérémie Galarneau <jeremie.galarneau@efficios.com>
 */

#ifndef BABELTRACE_CTF_WRITER_FUNCTOR_H
#define BABELTRACE_CTF_WRITER_FUNCTOR_H

#include <glib.h>
#include "common/macros.h"

void value_exists(gpointer element, gpointer search_query);

#endif /* BABELTRACE_CTF_WRITER_FUNCTOR_H */
