/*
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * Copyright (C) 2012 Yannick Brosseau <yannick.brosseau@gmail.com>
 * Copyright (C) 2016 Jérémie Galarneau <jeremie.galarneau@efficios.com>
 *
 * Lib BabelTrace - Common function to all tests
 */

#ifndef BABELTRACE_TESTS_UTILS_COMMON_H
#define BABELTRACE_TESTS_UTILS_COMMON_H

void recursive_rmdir(const char *path);

#endif /* BABELTRACE_TESTS_UTILS_COMMON_H */
