/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright (c) 2017 Philippe Proulx <pproulx@efficios.com>
 */

#ifndef BABELTRACE_CTF_WRITER_LOGGING_H
#define BABELTRACE_CTF_WRITER_LOGGING_H

#define BT_LOG_OUTPUT_LEVEL bt_lib_ctf_writer_log_level
#include "logging/log.h"

BT_LOG_LEVEL_EXTERN_SYMBOL(bt_lib_ctf_writer_log_level);

#endif /* BABELTRACE_CTF_WRITER_LOGGING_H */
