/*
 * Copyright (c) 2022-2024 Philippe Proulx <pproulx@efficios.com>
 *
 * SPDX-License-Identifier: MIT
 */

#ifndef BABELTRACE_PLUGINS_CTF_COMMON_METADATA_JSON_STRINGS_HPP
#define BABELTRACE_PLUGINS_CTF_COMMON_METADATA_JSON_STRINGS_HPP

namespace ctf {
namespace jsonstr {

extern const char * const accuracy;
extern const char * const align;
extern const char * const attrs;
extern const char * const bigEndian;
extern const char * const bitOrder;
extern const char * const btNs;
extern const char * const byteOrder;
extern const char * const clkCls;
extern const char * const cycles;
extern const char * const dataStreamCls;
extern const char * const dataStreamClsId;
extern const char * const dataStreamId;
extern const char * const defClkClsId;
extern const char * const defClkTs;
extern const char * const descr;
extern const char * const discEventRecordCounterSnap;
extern const char * const dynLenArray;
extern const char * const dynLenBlob;
extern const char * const dynLenStr;
extern const char * const elemFc;
extern const char * const emfUri;
extern const char * const encoding;
extern const char * const env;
extern const char * const eventRecordCls;
extern const char * const eventRecordClsId;
extern const char * const eventRecordCommonCtx;
extern const char * const eventRecordCommonCtxFc;
extern const char * const eventRecordHeader;
extern const char * const eventRecordHeaderFc;
extern const char * const eventRecordPayload;
extern const char * const eventRecordSpecCtx;
extern const char * const extensions;
extern const char * const fc;
extern const char * const fcAlias;
extern const char * const fixedLenBitArray;
extern const char * const fixedLenBitMap;
extern const char * const fixedLenBool;
extern const char * const fixedLenFloat;
extern const char * const fixedLenSInt;
extern const char * const fixedLenUInt;
extern const char * const flags;
extern const char * const freq;
extern const char * const ftl;
extern const char * const id;
extern const char * const len;
extern const char * const lenFieldLoc;
extern const char * const littleEndian;
extern const char * const logLevel;
extern const char * const logLevelAlert;
extern const char * const logLevelCritical;
extern const char * const logLevelDebugFunction;
extern const char * const logLevelDebugLine;
extern const char * const logLevelDebugModule;
extern const char * const logLevelDebug;
extern const char * const logLevelDebugProcess;
extern const char * const logLevelDebugProgram;
extern const char * const logLevelDebugSystem;
extern const char * const logLevelDebugUnit;
extern const char * const logLevelEmergency;
extern const char * const logLevelError;
extern const char * const logLevelInfo;
extern const char * const logLevelNotice;
extern const char * const logLevelWarning;
extern const char * const ltf;
extern const char * const mappings;
extern const char * const mediaType;
extern const char * const memberClasses;
extern const char * const metadataStreamUuid;
extern const char * const minAlign;
extern const char * const name;
extern const char * const ns;
extern const char * const nullTerminatedStr;
extern const char * const offsetFromOrigin;
extern const char * const optional;
extern const char * const opts;
extern const char * const origin;
extern const char * const path;
extern const char * const payloadFc;
extern const char * const pktContentLen;
extern const char * const pktCtx;
extern const char * const pktCtxFc;
extern const char * const pktEndDefClkTs;
extern const char * const pktHeader;
extern const char * const pktHeaderFc;
extern const char * const pktMagicNumber;
extern const char * const pktSeqNum;
extern const char * const pktTotalLen;
extern const char * const preamble;
extern const char * const precision;
extern const char * const prefDispBase;
extern const char * const roles;
extern const char * const seconds;
extern const char * const selFieldLoc;
extern const char * const selFieldRanges;
extern const char * const specCtxFc;
extern const char * const staticLenArray;
extern const char * const staticLenBlob;
extern const char * const staticLenStr;
extern const char * const structure;
extern const char * const traceCls;
extern const char * const type;
extern const char * const uid;
extern const char * const unixEpoch;
extern const char * const utf16Be;
extern const char * const utf16Le;
extern const char * const utf32Be;
extern const char * const utf32Le;
extern const char * const utf8;
extern const char * const uuid;
extern const char * const variant;
extern const char * const varLenSInt;
extern const char * const varLenUInt;
extern const char * const version;
extern const char * const version;

} /* namespace jsonstr */
} /* namespace ctf */

#endif /* BABELTRACE_PLUGINS_CTF_COMMON_METADATA_JSON_STRINGS_HPP */
