#!/bin/bash
#
# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (C) 2015 Antoine Busque <abusque@efficios.com>
# Copyright (C) 2019 Michael Jeanson <mjeanson@efficios.com>
#

if [ -n "${BT_TESTS_SRCDIR:-}" ]; then
	UTILSSH="$BT_TESTS_SRCDIR/utils/utils.sh"
else
	UTILSSH="$(dirname "$0")/../../utils/utils.sh"
fi

# shellcheck source=../../utils/utils.sh
source "$UTILSSH"

this_dir_relative="plugins/flt.lttng-utils.debug-info"
this_dir_build="$BT_TESTS_BUILDDIR/$this_dir_relative"
debug_info_data="$BT_TESTS_DATADIR/$this_dir_relative/powerpc64le-linux-gnu"

"$this_dir_build/test-bin-info" \
	--foo-addr=0x2e7c \
	--printf-offset=0x190 \
	--printf-lineno=36 \
	--tp-offset=0x1c \
	--tp-lineno=35 \
	--debug-link-crc=0x9b8eb2ff \
	--debug-info-dir "$debug_info_data" \
	--build-id cdd98cdd87f7fe64c13b6daad553987eafd40cbb
