/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright (C) 2010-2019 EfficiOS Inc. and Linux Foundation
 */

#ifndef BABELTRACE2_CTF_WRITER_TYPES_H
#define BABELTRACE2_CTF_WRITER_TYPES_H

#ifdef __cplusplus
extern "C" {
#endif

/// False boolean value for the #bt_ctf_bool type.
#define BT_CTF_FALSE    0

/// True boolean value for the #bt_ctf_bool type.
#define BT_CTF_TRUE     1

typedef int bt_ctf_bool;

#ifdef __cplusplus
}
#endif

#endif /* BABELTRACE2_CTF_WRITER_TYPES_H */
